//
// Program to generate 1H2D hierarchical (hierarchies in rows) cell
// suppression problem instances as a set of 2D tables 
// command line arguments: 
//        mnrows: mean number of rows of tables (randomly obtained from a 
//               uniform distribution [1/2 nrows, 3/2 nrows]).
//        ncols: number of columns for each table 
//        deep:  deep of hierarchies (deep of hierarchical tree).
//               deep=1 generates a single table.
//        minhr: minimum number of rows with hierarchies for each table,
//               i.e., sons of each table in the hierarchical tree
//        maxhr: maximum number of rows with hierarchies for each table, 
//               i.e., sons of each table in the hierarchical tree.
//               The real number of hierarchical rows is randomly obtained
//               from a uniform distribution [minhr,maxhr]
//        p:     number of primary suppressions (uniformly distributed among 
//               the tables, randomly located within each table)
//        p_l%:  lower level protection as a p_l% of the cell value
//        p_u%:  upper level protection as a p_u% of the cell value
//               (the number is randomly generated, from 2 to maxsub)
//        tweight: type of weights (1= value, 2= log value, 3=binary)
//        numb:  type of numbering, 0 or 1 (vectors start in 0 or 1)
//        seed:  integer initial seed for random numbers 
//                                                                      
// Values of internals cells are obtained from a uniform distribution
// between [1..N] with probability 0.8 and are 0 with probability 0.1.
// N is different for tables of different levels, and is adjusted to
// provide consistent values (i.e.,of similar order) for all the cells 
// of a table. Values of primaries come from a uniform distribution 
// [1..N/10].
//
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <vector>

using namespace std;

/* typedefs */
const int EMPTY= -1;
typedef struct{
    int m; // number of internal rows 
    int n; // number of internal columns 
    vector<double> v; /* (m+1)*(n+1) table values (from 0..m to 0..n,
		      marginals at row m, column n) */
    vector<double> w; /* (m+1)*(n+1) table weights(from 0..m to 0..n,
		      marginals at row m, column n) */
    vector<bool>is_prim; /* (m+1)*(n+1) boolean table: tells if each cell is
			    or not primary */
    vector<int> rowhier; //for each row gives: if hierarchical, the table
                         //where is decomposed, if not hierarchical, EMPTY
    int parent; //parent table of this table
    int row_in_parent; //row of parent table associated to 
                       //marginal row of this table 
} TABLE2D;
typedef struct{
    int p_t; //table
    int p_m; //row
    int p_n; //column
    int p_lpl; //lower protection level
    int p_upl; //upper protection level
} PRIMARYCELL;
typedef struct{
    vector<TABLE2D> vtab; // vector of 2D tables
    vector<PRIMARYCELL> vprim; //vector of primary cells
    vector<int> bfl; /* breadht-first-list of hierarchical tree
			it could be avoided since bfl[i]=i in
			this implementation */
} TABLES1H2D;

/* function prototypes */
void create_tables(TABLES1H2D &tab, int mnrows, int ncols, int deep, 
		   int minhr, int maxhr,
		   int p, double pl, double pu, int tweight);
void allocate_table(TABLE2D &tab2D, int rows, int cols);
double get_cell2D(TABLE2D &tab, int row, int column);
void put_cell2D(TABLE2D &tab, int row, int column, double value);
double get_weight2D(TABLE2D &tab, int row, int column);
void put_weight2D(TABLE2D &tab, int row, int column, double weight);
bool is_primary2D(TABLE2D &tab, int row, int column);
void set_primary2D(TABLE2D &tab, int row, int column, bool b);
int table_of_hierarchical_row2D(TABLE2D &tab, int row);
void compute_marginals(TABLE2D &tab);
void write_table_AMPL_format(TABLES1H2D &tab, int mnrows, int ncols,int deep,
			     int minhr, int maxhr, int p, double pl, double pu,
			     int tweight, int numb, int seed);
double uniform(double min, double max);

int main(int argc, char *argv[])
{
  
    int mnrows, ncols, deep, minhr, maxhr, p, tweight, numb;
    int seed;
    TABLES1H2D tab;
    double pl,pu;
    
    /* check command line arguments */
    
    if (argc != 12) {
	fprintf(stderr,"usage: gen_CSP_instance_1H2D mnrows ncols deep minhr maxhr p p_l%% p_u%% tweight numb seed\n");
	exit(-1);
    }
    
    sscanf(argv[1],"%d",&mnrows);
    if (mnrows < 1) {
	fprintf(stderr,"error: mnrows must be greater than 0 \n");
	exit(-1);
    }
    sscanf(argv[2],"%d",&ncols);
    if (ncols < 1) {
	fprintf(stderr,"error: ncols must be greater than 0 \n");
	exit(-1);
    }
    sscanf(argv[3],"%d",&deep);
    if (deep < 1) {
	fprintf(stderr,"error: deep must be greater than 0 \n");
	exit(-1);
    }
    sscanf(argv[4],"%d",&minhr);
    if (minhr < 0) {
	fprintf(stderr,"error: minhr must be greater or equal than 0 \n");
	exit(-1);
    }
    sscanf(argv[5],"%d",&maxhr);
    if (maxhr < minhr) {
	fprintf(stderr,"error: maxhr must be greater or equal than minhr \n");
	exit(-1);
    }
    sscanf(argv[6],"%d",&p);
    if (p < 1) {
	fprintf(stderr,"error: p must be greater than 0 \n");
	exit(-1);
    }
    sscanf(argv[7],"%lf",&pl);
    if (pl <= 0.00000) {
	fprintf(stderr,"error: p_l%% must be >0.0 \n");
	exit(-1);
    }
    sscanf(argv[8],"%lf",&pu);
    if (pu <= 0.00000) {
	fprintf(stderr,"error: p_u%% must be >0.0 \n");
	exit(-1);
    }
    sscanf(argv[9],"%d",&tweight);
    if (tweight >3 || tweight < 1) {
	fprintf(stderr,"error: tweight must be 1,2 or 3 \n");
	exit(-1);
    }
    sscanf(argv[10],"%d",&numb);
    if (numb != 0  && numb != 1) {
	fprintf(stderr,"error: numb must be 0 or 1\n");
	exit(-1);
    }

    /* set random generator seed */
    sscanf(argv[11],"%d",&seed);
    srand((unsigned int)seed);

    /* create hierarchical table structure */
    create_tables(tab,mnrows,ncols,deep,minhr,maxhr,p,pl,pu,tweight);
  
    /* write_table in AMPL format */
    write_table_AMPL_format(tab,mnrows,ncols,deep,minhr,maxhr,p,pl,pu,
			    tweight,numb,seed);

    return(0);
}

void create_tables(TABLES1H2D &tab, int mnrows, int ncols, int deep, 
		   int minhr, int maxhr,
		   int p, double pl, double pu, int tweight)
{
    
    
    tab.vprim.resize(p);
    
    //create hierarchical structure of tables
    int nt=1; //current number of tables 
    tab.bfl.resize(nt);
    tab.bfl[0]= 0;
    int ini_bfl=0;//index of first table in breadth-first list for current deep
    int fi_bfl=0; //index of last table in breadth-first list for current deep
    tab.vtab.resize(nt);
    allocate_table(tab.vtab[0],(int)uniform(((double)mnrows/2.0),
					    (3.0*(double)mnrows/2.0)),
		   ncols);
    int i;
    for(i=0;i<deep;i++) {
	for(int j=ini_bfl;j<=fi_bfl;j++) {
	    int currt= tab.bfl[j]; //current table
	    int nhr= (int)uniform((double)minhr,(double)maxhr);
	    if (nhr>tab.vtab[currt].m) nhr= tab.vtab[currt].m; 
	    tab.bfl.resize(nt+nhr);
	    tab.vtab.resize(nt+nhr);
	    for(int l=0;l<nhr;l++){
		allocate_table(tab.vtab[nt],
			       (int)uniform((double)mnrows/2.0,
					    3.0*mnrows/2.0),ncols);
		tab.vtab[nt].parent= currt;
		// generate position of hierarchical row, detecting duplicates
		//only internal rows
		int rh= (int)uniform(0.0,(double)tab.vtab[currt].m-1);
		while (tab.vtab[currt].rowhier[rh] != EMPTY)
		    rh= (rh+1)%(tab.vtab[currt].m);//only internal rows
		tab.vtab[nt].row_in_parent= rh;
		tab.vtab[currt].rowhier[rh]= nt;
		tab.bfl[nt]= nt;
		nt++;
	    }
	}
	ini_bfl= fi_bfl+1;
	fi_bfl= nt-1;
    }

    //fill internal cells of tables dealing with hierarchical rows
    //tables are scanned bottom-up

    //first compute number of primaries per table
    int total_cells= 0;
    int t;
    for(t=0;t<(int)tab.vtab.size();t++) 
	total_cells += (tab.vtab[t].m+1)*(tab.vtab[t].n+1);

    vector<int> np_per_tab(tab.vtab.size(),0);
    int acum_p=0;
    for(t=(int)tab.vtab.size()-1;t>0;t--) {
	int p_tab= (int)((double)(tab.vtab[t].m+1)*(tab.vtab[t].n+1)/
			 (double)total_cells*(double)p);
	np_per_tab[t]= p_tab;
	acum_p += p_tab;
    }
    np_per_tab[0]= p-acum_p;



    //fill tables 
    int ip=0; //index of current primary cell
    for(int h= (int)tab.bfl.size()-1;h>=0;h--) {
	t= tab.bfl[h];
	int deep_t= 0;
	for(i=t;i!=EMPTY;i=tab.vtab[i].parent) deep_t++;
	double N= 100*pow(mnrows,deep-deep_t+1);
	//internal cells
	for(i=0;i<tab.vtab[t].m;i++) 
	    if (tab.vtab[t].rowhier[i]==EMPTY) { //not yet filled if EMPTY
		for(int j=0;j<tab.vtab[t].n;j++) {
		    double value;
		    if (uniform(0.0,1.0) > 0.9) //10% cells are 0
			value= 0.0; 
		    else
			value= ceil(uniform(1.0,N));
		    put_cell2D(tab.vtab[t],i,j,value);
		}
	    }
	//primary cells
	for(i=0;i<np_per_tab[t];i++) {
	    //location
	    int r= (int)uniform(0.0,(double)tab.vtab[t].m-1);
	    int c= (int)uniform(0.0,(double)tab.vtab[t].n-1);
	    //check duplicates and zero cells
	    while (is_primary2D(tab.vtab[t],r,c) || 
		   get_cell2D(tab.vtab[t],r,c)==0.0) {
		c= (c+1)%(tab.vtab[t].n); //only internal columns
		if (c==0) r= (r+1)%(tab.vtab[t].m); //only internal rows
	    }
	    //value if not already set because is hierarchical row
	    double value;
	    int table_hier_row= table_of_hierarchical_row2D(tab.vtab[t],r);
	    if (table_hier_row==EMPTY) {
		value= ceil(uniform(1.0,N/10));
		put_cell2D(tab.vtab[t],r,c,value);
	    }
	    else {
		value= get_cell2D(tab.vtab[table_hier_row],
				  tab.vtab[table_hier_row].m,c);
	    }
	    //store primary in current table 
	    tab.vprim[ip].p_t= t;
	    tab.vprim[ip].p_m= r;
	    tab.vprim[ip].p_n= c;
	    tab.vprim[ip].p_lpl= (int)ceil(pl/100.0*value);;
	    tab.vprim[ip].p_upl= (int)ceil(pu/100.0*value);
	    set_primary2D(tab.vtab[t],r,c,true);
	    ip++;
	    //store primary also in marginal row of subtable if 
	    //hierarchical row; indeed, this is an "artificial primary"
	    if (table_hier_row!=EMPTY) {
		//extra space for the extra artificial primary
		tab.vprim.resize(tab.vprim.size()+1);  
		tab.vprim[ip].p_t= table_hier_row;
		tab.vprim[ip].p_m= tab.vtab[table_hier_row].m;
		tab.vprim[ip].p_n= c;
		tab.vprim[ip].p_lpl= (int)ceil(pl/100.0*value);;
		tab.vprim[ip].p_upl= (int)ceil(pu/100.0*value);
		set_primary2D(tab.vtab[table_hier_row],
			      tab.vtab[table_hier_row].m,c,true);
		ip++;
	    }
	}
	//compute marginals
	compute_marginals(tab.vtab[t]);
	
	//compute weights
	for(i=0;i<tab.vtab[t].m+1;i++)
            for(int j=0;j<tab.vtab[t].n+1;j++) {
                double value= get_cell2D(tab.vtab[t],i,j);
                double weight;
                if (tweight==1) //value
                    weight= value;
                else if (tweight==2) //log value
                    if (value<1.0)
                        weight= 0.0;
                    else
                        weight= log10(value);
                else // bin
                    weight= 1.0;
                put_weight2D(tab.vtab[t],i,j,weight);
            }
	
	//copy row marginal in parent table if needed
	int taux= tab.vtab[t].parent;
	if (taux != EMPTY) {
	    int raux= tab.vtab[t].row_in_parent;
	    for(int j=0;j<tab.vtab[t].n;j++){
		put_cell2D(tab.vtab[taux],raux,j,
			   get_cell2D(tab.vtab[t],tab.vtab[t].m,j));
	    }
	}
    }
}

// allocates a single 2D table
void allocate_table(TABLE2D &tab2D, int rows, int cols)
{
    tab2D.m= rows;
    tab2D.n= cols;
    tab2D.v.resize((rows+1)*(cols+1));
    tab2D.w.resize((rows+1)*(cols+1));
    tab2D.is_prim.resize((rows+1)*(cols+1),false);
    tab2D.rowhier.resize(rows,EMPTY);
    tab2D.parent= EMPTY;
    tab2D.row_in_parent= EMPTY;
}

// randomly generate a value of the real uniform distribution [min,max]
double uniform(double min, double max)
{
    return((min + (max-min)*(((double)rand())/((double)RAND_MAX))));
}

inline double get_cell2D(TABLE2D &tab, int row, int column)
{
    return(tab.v[row*(tab.n+1)+column]);
}

inline void put_cell2D(TABLE2D &tab, int row, int column, double value)
{
    tab.v[row*(tab.n+1)+column]= value;
}

inline double get_weight2D(TABLE2D &tab, int row, int column)
{
    return(tab.w[row*(tab.n+1)+column]);
}

inline void put_weight2D(TABLE2D &tab, int row, int column, double weight)
{
    tab.w[row*(tab.n+1)+column]= weight;
}

inline bool is_primary2D(TABLE2D &tab, int row, int column)
{
    return(tab.is_prim[row*(tab.n+1)+column]);
}

inline void set_primary2D(TABLE2D &tab, int row, int column, bool b)
{
    tab.is_prim[row*(tab.n+1)+column]= b;
}

inline int table_of_hierarchical_row2D(TABLE2D &tab, int row)
{
    return(tab.rowhier[row]);
}

void compute_marginals(TABLE2D &tab)
{
    int i,j;
    double *totc,totr,totcheck,tot;
  
    totc= new double[tab.n];
    for(j=0;j<tab.n;j++) totc[j]= 0;
    tot= 0.0;
    for(i=0;i<tab.m;i++) {
	totr= 0.0;
	for(j=0;j<tab.n;j++) {
	    totc[j] += get_cell2D(tab,i,j);
	    totr += get_cell2D(tab,i,j);
	}
	put_cell2D(tab,i,tab.n,totr);
	tot= tot+totr;
    }
    totcheck= 0.0;
    for(j=0;j<tab.n;j++) {
	put_cell2D(tab,tab.m,j,totc[j]);
	totcheck += totc[j];
    }
    if (tot != totcheck) {
	fprintf(stderr,"compute_marginals: internal error, sum of row marginals != sum of column marginals %f %f\n",tot,totcheck);
	exit(-1);
    }
    put_cell2D(tab,tab.m,tab.n,tot);
    delete[]totc;
}


// write to standard_output
// in AMPL format indexes start in 1, not in 0
void write_table_AMPL_format(TABLES1H2D &tab, int mnrows, int ncols,int deep,
			     int minhr, int maxhr, int p, double pl, double pu,
			     int tweight, int numb, int seed)
{

    printf("# generated with gen_CSP_instance_1H2D using parameters\n");
    printf("# \t\t %d %d %d %d %d %d %f %f %d %d %d\n", 
	   mnrows,ncols,deep,minhr,maxhr,p,pl,pu,tweight,numb,seed);

    printf("#T= number of 2D tables\n");
    printf("param T := %d;\n",tab.vtab.size());

    printf("#M= number of rows of each table (without marginal row)\n");
    printf("param M := \n");
    int i;
    for(i=0;i<(int)tab.vtab.size();i++)
        printf("%d \t %d \n",i+numb,tab.vtab[i].m);
    printf(";\n");

    printf("#N= number of columns common to all the tables (without marginal column))\n");
    printf("param N := %d;\n",tab.vtab[0].n);

    printf("#(T-1) lines with hierarchical info: row rh of table th decomposed in table tdh\n");
    printf("param: \t rh \t th \t tdh :=\n");
    for(i=1;i<(int)tab.vtab.size();i++) //table 0 is parent
	if (tab.vtab[i].parent != EMPTY) printf("%d \t %d \t %d \t %d\n",
						i-1+numb,
						tab.vtab[i].row_in_parent+numb,
						tab.vtab[i].parent+numb,
						i+numb);
    printf(";\n");

    printf("#P= number of primary suppression cells\n");
    printf("param P := %d;\n",tab.vprim.size());

    printf("#for each primary: table, row, column, lower prot., and upper prot.\n");
    printf("param: \t p_t \t p_r \t p_c \t lpl \t upl :=\n");
    for(i=0;i<(int)tab.vprim.size();i++)
        printf("%d \t %d \t %d \t %d \t %d \t %d\n",i+numb,
	       tab.vprim[i].p_t+numb,
	       tab.vprim[i].p_m+numb,
	       tab.vprim[i].p_n+numb,
	       tab.vprim[i].p_lpl,tab.vprim[i].p_upl);
    printf(";\n");

    printf("#table values and weights (indexed by (table,row,column)), marginals included\n");
    printf("param: \t\t\t a \t weight :=\n");
    for(unsigned int t=0;t<tab.vtab.size();t++)
        for(int i=0;i<=tab.vtab[t].m;i++)
            for(int j=0;j<=tab.vtab[t].n;j++)
                printf("%d \t %d\t %d \t %f \t %f\n",t+numb,i+numb,j+numb,
                       get_cell2D(tab.vtab[t],i,j),
		       get_weight2D(tab.vtab[t],i,j));
    printf(";\n");
}
